/*
 * Copyright (c) 2006-2008 Amit Singh/Google Inc.
 * Copyright (c) 2010 Tuxera Inc.
 * Copyright (c) 2012-2022 Benjamin Fleischer
 * All rights reserved.
 */

#ifndef _FUSE_LOCKING_H_
#define _FUSE_LOCKING_H_

#include "fuse.h"

#include "fuse_node.h"

#include <fuse_param.h>

#include <kern/thread.h>
#include <libkern/locks.h>

#ifdef FUSE_TRACE_LK
#  include <sys/vm.h>
#endif

enum fusefslocktype {
    FUSEFS_SHARED_LOCK    = 1,
    FUSEFS_EXCLUSIVE_LOCK = 2,
    FUSEFS_FORCE_LOCK     = 3
};

#if __LP64__
#  define FUSEFS_SHARED_OWNER (void *)0xffffffffffffffff
#else
#  define FUSEFS_SHARED_OWNER (void *)0xffffffff
#endif

/* Locking */
extern int fusefs_lock(fusenode_t, enum fusefslocktype);
extern bool fusefs_lock_upgrade(fusenode_t cp);
extern int fusefs_lockpair(fusenode_t, fusenode_t, enum fusefslocktype);
extern int fusefs_lockfour(fusenode_t, fusenode_t, fusenode_t, fusenode_t,
                           enum fusefslocktype);
extern void fusefs_lock_truncate(fusenode_t, lck_rw_type_t);

/* Unlocking */
extern void fusefs_unlock(fusenode_t);
extern void fusefs_unlockpair(fusenode_t, fusenode_t);
extern void fusefs_unlockfour(fusenode_t, fusenode_t, fusenode_t, fusenode_t);
extern void fusefs_unlock_truncate(fusenode_t);

extern lck_attr_t     *fuse_lock_attr;
extern lck_grp_attr_t *fuse_group_attr;
extern lck_grp_t      *fuse_lock_group;
extern lck_mtx_t      *fuse_device_mutex;

#ifdef FUSE_TRACE_LK

#define fuse_lck_mtx_lock(m)                                                  \
    {                                                                         \
        proc_t __FUNCTION__ ## p = current_proc();                            \
        IOLog("0: lck_mtx_lock(%p): %s@%d by %d\n", (m), __FUNCTION__,        \
           __LINE__, (__FUNCTION__ ## p) ? proc_pid(__FUNCTION__ ## p) : 0);  \
        lck_mtx_lock((m));                                                    \
        IOLog("1: lck_mtx_lock(%p): %s@%d by %d\n", (m), __FUNCTION__,        \
           __LINE__, (__FUNCTION__ ## p) ? proc_pid(__FUNCTION__ ## p) : 0);  \
    }

#define fuse_lck_mtx_unlock(m)                                                \
    {                                                                         \
        proc_t __FUNCTION__ ## p = current_proc();                            \
        IOLog("1: lck_mtx_unlock(%p): %s@%d by %d\n", (m), __FUNCTION__,      \
           __LINE__, (__FUNCTION__ ## p) ? proc_pid(__FUNCTION__ ## p) : 0);  \
        lck_mtx_unlock((m));                                                  \
        IOLog("0: lck_mtx_unlock(%p): %s@%d by %d\n", (m), __FUNCTION__,      \
           __LINE__, (__FUNCTION__ ## p) ? proc_pid(__FUNCTION__ ## p) : 0);  \
    }

#define fuse_lck_rw_lock_shared(l)      lck_rw_lock_shared((l))
#define fuse_lck_rw_lock_exclusive(l)   lck_rw_lock_exclusive((l))
#define fuse_lck_rw_unlock_shared(l)    lck_rw_unlock_shared((l))
#define fuse_lck_rw_unlock_exclusive(l) lck_rw_unlock_exclusive((l))

#else /* !FUSE_TRACE_LK */

#define fuse_lck_mtx_lock(m)            lck_mtx_lock((m))
#define fuse_lck_mtx_unlock(m)          lck_mtx_unlock((m))

#define fuse_lck_rw_lock_shared(l)      lck_rw_lock_shared((l))
#define fuse_lck_rw_lock_exclusive(l)   lck_rw_lock_exclusive((l))
#define fuse_lck_rw_unlock_shared(l)    lck_rw_unlock_shared((l))
#define fuse_lck_rw_unlock_exclusive(l) lck_rw_unlock_exclusive((l))

#define fuse_lck_mtx_try_lock(l)        IOLockTryLock((IOLock *)l)

#endif /* FUSE_TRACE_LK */

typedef struct _fusefs_recursive_lock fusefs_recursive_lock;

extern fusefs_recursive_lock* fusefs_recursive_lock_alloc(void);
extern fusefs_recursive_lock* fusefs_recursive_lock_alloc_with_maxcount(UInt32);
extern void fusefs_recursive_lock_free(fusefs_recursive_lock *lock);
extern void fusefs_recursive_lock_lock(fusefs_recursive_lock *lock);
extern void fusefs_recursive_lock_unlock(fusefs_recursive_lock *lock);
extern bool fusefs_recursive_lock_have_lock(fusefs_recursive_lock *lock);

#if M_OSXFUSE_ENABLE_LOCK_LOGGING

extern lck_mtx_t *fuse_log_lock;

#define rawlog(msg, args...) IOLog(msg, ##args)

#define log(fmt, args...) \
    do { \
        lck_mtx_lock(fuse_log_lock); \
        rawlog(fmt, ##args); \
        rawlog("\n"); \
        lck_mtx_unlock(fuse_log_lock); \
    } while(0)

#define log_enter(params_format, args...) \
    do { \
        lck_mtx_lock(fuse_log_lock); \
        rawlog("[%s:%d] Entering %s: ", __FILE__, __LINE__, __FUNCTION__); \
        rawlog(params_format, ##args); \
        rawlog("\n"); \
        lck_mtx_unlock(fuse_log_lock); \
    } while(0)

#define log_leave(return_format, args...) \
    do { \
        lck_mtx_lock(fuse_log_lock); \
        rawlog("[%s:%d] Leaving %s: ", __FILE__, __LINE__, __FUNCTION__); \
        rawlog(return_format, ##args); \
        rawlog("\n"); \
        lck_mtx_unlock(fuse_log_lock); \
    } while(0)

#else /* !M_OSXFUSE_ENABLE_LOCK_LOGGING */

#define log(fmt, args...) do {} while(0)
#define log_enter(params_format, args...) do {} while(0)
#define log_leave(return_format, args...) do {} while(0)

#endif /* M_OSXFUSE_ENABLE_LOCK_LOGGING */

typedef fusefs_recursive_lock fuse_biglock_t;

#define fuse_biglock_alloc() fusefs_recursive_lock_alloc_with_maxcount(1)
#define fuse_biglock_free(lock) fusefs_recursive_lock_free(lock)

#define fuse_biglock_lock(lock) \
    do { \
        thread_t fbl_thread __unused = current_thread(); \
        log("%s thread=" FUSE_LOG_P ": locking biglock " FUSE_LOG_P " ...", \
            __FUNCTION__, FUSE_LOG_P_MASK(fbl_thread), FUSE_LOG_P_MASK(lock)); \
        fusefs_recursive_lock_lock(lock); \
        log("%s thread=" FUSE_LOG_P ": biglock " FUSE_LOG_P " locked", \
            __FUNCTION__, FUSE_LOG_P_MASK(fbl_thread), FUSE_LOG_P_MASK(lock)); \
    } while(0)

#define fuse_biglock_unlock(lock) \
    do { \
        thread_t fbu_thread __unused = current_thread(); \
        log("%s thread=" FUSE_LOG_P ": unlocking biglock " FUSE_LOG_P " ...", \
            __FUNCTION__, FUSE_LOG_P_MASK(fbu_thread), FUSE_LOG_P_MASK(lock)); \
        fusefs_recursive_lock_unlock(lock); \
        log("%s thread=" FUSE_LOG_P ": biglock " FUSE_LOG_P " unlocked", \
            __FUNCTION__, FUSE_LOG_P_MASK(fbu_thread), FUSE_LOG_P_MASK(lock)); \
    } while(0)

#define fuse_biglock_have_lock(lock) fusefs_recursive_lock_have_lock(lock)

#endif /* _FUSE_LOCKING_H_ */
