//
//  Helper.c
//  Helper
//
//  Created by Benjamin Fleischer on 02.12.12.
//  Copyright (c) 2012-2020 Benjamin Fleischer. All rights reserved.
//

#include <AvailabilityMacros.h>
#include <err.h>
#include <fts.h>
#include <libkern/OSKextLib.h>
#include <netinet/in.h>
#include <stdio.h>
#include <sys/socket.h>
#include <sys/stat.h>
#include <sys/sysctl.h>
#include <sys/types.h>
#include <sysexits.h>
#include <unistd.h>

#include <CoreServices/CoreServices.h>
#include <Foundation/Foundation.h>
#include <IOKit/kext/KextManager.h>
#include <Security/Security.h>

#include "fuse_param.h"
#include "fuse_version.h"

#include "MounterCommon.h"
#include "BetterAuthorizationSampleLib.h"

static Boolean CheckBundleSignature(CFURLRef bundleURL, aslclient asl, aslmsg aslMsg)
{
    Boolean retval = true;

    int systemVersionMajor;
    int systemVersionMinor;

    if (!GetSystemVersion(&systemVersionMajor, &systemVersionMinor, NULL)) {
        asl_log(asl, aslMsg, ASL_LEVEL_DEBUG, "Failed to determine system version");

        // Assume we are runing on Mac OS X 10.6 or later
        systemVersionMajor = 10;
        systemVersionMinor = 6;
    }

    /*
     * Note: Although the CodeSigning API is available as private API on
     * Mac OS X 10.5 (Intel) we cannot use it because it behaves differently
     * than on later versions of Mac OS X.
     */
    if (systemVersionMajor > 10 || (systemVersionMajor == 10 && systemVersionMinor >= 6)) {
        OSStatus status;
        CFStringRef requirementText = CFSTR("identifier \"" OSXFUSE_BUNDLE_IDENTIFIER "\" and "
                                            "anchor apple generic and"
                                            "(certificate leaf[subject.OU] = \"" OSXFUSE_CODESIGN_SUBJECT_OU "\" or "
                                            " certificate leaf[subject.OU] = \"" OSXFUSE_CODESIGN_DEV_SUBJECT_OU "\")");
        SecStaticCodeRef staticCode = NULL;
        SecRequirementRef requirement = NULL;

        retval = false;

        if(!&SecStaticCodeCheckValidity) {
            // We should never get here
            goto out;
        }

        status = SecStaticCodeCreateWithPath(bundleURL, kSecCSDefaultFlags, &staticCode);
        if (status != errSecSuccess) {
            asl_log(asl, aslMsg, ASL_LEVEL_DEBUG, "Failed to validate bundle signature: Create a static code");
            goto out;
        }

        // Create signature requirement
        status = SecRequirementCreateWithString(requirementText, kSecCSDefaultFlags, &requirement);
        if (status != errSecSuccess) {
            asl_log(asl, aslMsg, ASL_LEVEL_DEBUG, "Failed to validate bundle signature: Create a requirement (%ld)", (long)status);
            goto out;
        }

        // Check validity of signature
        status = SecStaticCodeCheckValidity(staticCode, kSecCSDefaultFlags, requirement);
        if (status != errSecSuccess) {
            asl_log(asl, aslMsg, ASL_LEVEL_DEBUG, "Failed to validate bundle signature: Check the static code validity");
            goto out;
        }

        // Bundle signature is valid
        retval = true;

out:
        if (staticCode)  CFRelease(staticCode);
        if (requirement) CFRelease(requirement);
    }

    return retval;
}

int PrepareKext(CFURLRef kextURL)
{
    int retval = 0;

    char kextPathC[MAXPATHLEN];
    CFBundleRef kextBundle = NULL;
    CFURLRef kextExecutableURL = NULL;
    char kextExecutablePathC[MAXPATHLEN];

    char *fts_path_argv[2];
    FTS *fts;
	FTSENT *fts_entry;
    mode_t mode;

    assert(kextURL != NULL);

    if (!CFURLGetFileSystemRepresentation(kextURL, true, (UInt8 *)kextPathC, sizeof(kextPathC))) {
        retval = ENAMETOOLONG;
        goto out;
    }
    fts_path_argv[0] = kextPathC;
    fts_path_argv[1] = NULL;

    // Change owner and mode of bundle
    if ((fts = fts_open(fts_path_argv, FTS_NOSTAT | FTS_PHYSICAL | FTS_XDEV, NULL)) == NULL) {
		retval = errno;
        goto out;
    }
	while ((fts_entry = fts_read(fts)) != NULL) {
        switch (fts_entry->fts_info) {
            case FTS_D:
                mode = 0755;
                break;

            case FTS_F:
            case FTS_NSOK:
            case FTS_SL:
            case FTS_SLNONE:
                mode = 0644;
                break;

            default:
                continue;
		}

        lchown(fts_entry->fts_path, 0, 0);
        chmod(fts_entry->fts_path, mode);
    }

    // Make bundle executable executable
    kextBundle = CFBundleCreate(NULL, kextURL);
    if (!kextBundle) {
        retval = EINVAL;
        goto out;
    }
    kextExecutableURL = CFBundleCopyExecutableURL(kextBundle);
    if (!kextExecutableURL) {
        retval = ENOENT;
        goto out;
    }
    if (!CFURLGetFileSystemRepresentation(kextExecutableURL, true, (UInt8 *)kextExecutablePathC, sizeof(kextExecutablePathC))) {
        retval = ENAMETOOLONG;
        goto out;
    }
    chmod(kextExecutablePathC, 0755);

out:
    if (kextBundle)         CFRelease(kextBundle);
    if (kextExecutableURL)  CFRelease(kextExecutableURL);

    return retval;
}

#pragma mark Command Handlers

static OSStatus GetVersionHandler(AuthorizationRef        auth,
                                  const void             *userData,
                                  CFDictionaryRef         request,
                                  CFMutableDictionaryRef  response,
                                  aslclient               asl,
                                  aslmsg                  aslMsg)
{
    int retval = 0;

    CFBundleRef bundle;

    CFStringRef versionString;
    CFNumberRef version;

    assert(auth != NULL);
	assert(request != NULL);
	assert(response != NULL);

#if 0
    {
        CFNumberRef commandVersion;
        double v;

        commandVersion = CFDictionaryGetValue(request, CFSTR(kBASCommandVersionKey));
        CFNumberGetValue(commandVersion, kCFNumberDoubleType, &v);

        asl_log(asl, aslMsg, ASL_LEVEL_DEBUG, "Command version: %f", v);
    }
#endif

    // Get helper bundle
    bundle = CFBundleGetMainBundle();
    assert(bundle != NULL);

    // Get helper version from bundle
    versionString = CFBundleGetValueForInfoDictionaryKey(bundle, CFSTR("CFBundleVersion"));
    assert(CFGetTypeID(versionString) == CFStringGetTypeID());
    {
        double v = CFStringGetDoubleValue(versionString);
        version = CFNumberCreate(kCFAllocatorDefault, kCFNumberDoubleType, &v);
    }

    CFDictionaryAddValue(response, CFSTR(kGetVersionResponseKeyVersion), version);

    CFRelease(version);

out:
    return BASErrnoToOSStatus(retval);
}

static OSStatus LoadKextHandler(AuthorizationRef        auth,
                                const void             *userData,
                                CFDictionaryRef         request,
                                CFMutableDictionaryRef  response,
                                aslclient               asl,
                                aslmsg                  aslMsg)
{
    int retval = 0;

    CFStringRef kextPath = NULL;
    char kextPathC[MAXPATHLEN];
    CFURLRef kextURL = NULL;

    assert(auth != NULL);
	assert(request != NULL);
	assert(response != NULL);

    if (!CFDictionaryGetValueIfPresent(request, CFSTR(kLoadKextRequestKeyPath), (const void **)&kextPath)) {
        retval = EINVAL;
        goto out;
    }
    kextURL = CFURLCreateWithFileSystemPath(kCFAllocatorDefault, kextPath, kCFURLPOSIXPathStyle, true);

    if (!CFStringGetFileSystemRepresentation(kextPath, kextPathC, sizeof(kextPathC))) {
        retval = ENAMETOOLONG;
        goto out;
    }
    asl_log(asl, aslMsg, ASL_LEVEL_DEBUG, "Loading kernel extension: %s", kextPathC);

    if (@available(macOS 10.10, *)) {
        if (faccessat(AT_FDCWD, kextPathC, F_OK, AT_EACCESS) != 0) {
            /*
             * The kernel extension for the currently running version of macOS
             * could not be found. Return ENOENT in addition to the path.
             */
            retval = ENOENT;
            goto out;
        }
    } else if (access(kextPathC, F_OK) != 0) {
        // See above
        retval = ENOENT;
        goto out;
    }

    // Check bundle signature
    if (!CheckBundleSignature(kextURL, asl, aslMsg)) {
        retval = EINVAL;
        goto out;
    }

    // Copy kernel extension to temporary location
    if (@available(macOS 12, *)) {
        @autoreleasepool {
            NSURL *tempDirURL = [NSURL fileURLWithPath:NSTemporaryDirectory()];
            NSURL *helperTempDirURL =  [tempDirURL URLByAppendingPathComponent:(NSString *)CFSTR(OSXFUSE_IDENTIFIER)];
            if (![NSFileManager.defaultManager createDirectoryAtURL:helperTempDirURL withIntermediateDirectories:YES attributes:nil error:nil]) {
                retval = ENOSPC;
                goto out;
            }

            NSURL *tempKextURL = [helperTempDirURL URLByAppendingPathComponent:(NSString *)CFSTR(OSXFUSE_KEXT_NAME)];
            if ([NSFileManager.defaultManager fileExistsAtPath:tempKextURL.path] &&
                ![NSFileManager.defaultManager removeItemAtURL:tempKextURL error:nil]) {
                retval = ENOSPC;
                goto out;
            }

            if (![NSFileManager.defaultManager copyItemAtURL:(NSURL *)kextURL toURL:tempKextURL error:nil]) {
                retval = ENOSPC;
                goto out;
            }

            CFRelease(kextURL);
            kextURL = (CFURLRef)CFBridgingRetain(tempKextURL);
        }
    }

    // Change owner and mode of kernel extension
    PrepareKext(kextURL);

    /*
     * Check bundle signature again before loading. This ensures that the kernel
     * extension has not been replaced or tampered with after performing the first
     * signature check. At this point admin rights are required to tamper with the
     * kernel extension.
     */
    if (!CheckBundleSignature(kextURL, asl, aslMsg)) {
        retval = EINVAL;
        goto out;
    }

    // Use KextManager to load kernel extension
    retval = KextManagerLoadKextWithURL(kextURL, NULL);
    
    if (retval == kOSReturnSuccess) {
        retval = 0;
    } else if (retval == kOSKextReturnNotFound) {
        retval = ENOENT;
    } else if (retval == kOSKextReturnSystemPolicy || retval == kOSKextReturnNotLoadable) {
        retval = EPERM;
    } else {
        retval = -1;
    }

out:
    if (kextURL) CFRelease(kextURL);

    return BASErrnoToOSStatus(retval);
}

static OSStatus UnloadKextHandler(AuthorizationRef        auth,
                                  const void             *userData,
                                  CFDictionaryRef         request,
                                  CFMutableDictionaryRef  response,
                                  aslclient               asl,
                                  aslmsg                  aslMsg)
{
    int retval = 0;

    assert(auth != NULL);
	assert(request != NULL);
	assert(response != NULL);

    if (@available(macOS 11, *)) {
        /*
         * KextManagerUnloadKextWithIdentifier() seems to be broken in macOS
         * 11.0, claiming the helper process, that is running as root, does not
         * have sufficient permissions (-603947004) to unload the kernel
         * extension. As a workaround we call the kmutil command line utility.
         */
        int pid = fork();
        if (pid == -1) {
            retval = -1;
        } else if (pid == 0) {
            (void)execl("/usr/bin/kmutil", "/usr/bin/kmutil", "unload", "-b", OSXFUSE_BUNDLE_IDENTIFIER, NULL);
            _exit(1);
        } else {
            int status = 0;
            if (waitpid(pid, &status, 0) == pid && WIFEXITED(status) && WEXITSTATUS(status) == 0) {
                retval = 0;
            } else {
                retval = -1;
            }
        }
    } else {
        // Use KextManager to unload kernel extension
        retval = KextManagerUnloadKextWithIdentifier(CFSTR(OSXFUSE_BUNDLE_IDENTIFIER));
        if (retval == kOSReturnSuccess) {
            retval = 0;
        } else {
            retval = -1;
        }
    }

out:
    return BASErrnoToOSStatus(retval);
}

static OSStatus LoadSysctlHandler(AuthorizationRef        auth,
                                  const void             *userData,
                                  CFDictionaryRef         request,
                                  CFMutableDictionaryRef  response,
                                  aslclient               asl,
                                  aslmsg                  aslMsg)
{
    int retval = 0;

    CFDictionaryRef sysctls;

    int i;
    char name[1024];

    CFIndex count;
    CFTypeRef *keys;
    CFTypeRef *values;
    CFStringRef key;
    CFDataRef value;

    if (!CFDictionaryGetValueIfPresent(request, CFSTR(kSetSysctlsRequestKeySysctls), (const void **)&sysctls)) {
        retval = EINVAL;
        goto out;
    }

    count = CFDictionaryGetCount(sysctls);

    keys = calloc(count, sizeof(CFTypeRef));
    if (!keys) {
        retval = errno;
        goto out;
    }
    values = calloc(count, sizeof(CFTypeRef));
    if (!values) {
        retval = errno;
        goto out;
    }
    CFDictionaryGetKeysAndValues(sysctls, (const void **)keys, (const void **)values);

    for (i = 0; i < count; i++) {
        if (CFGetTypeID(keys[i]) != CFStringGetTypeID() ||
            CFGetTypeID(values[i]) != CFDataGetTypeID()) {
            // Skip key-value pairs with illegal types
            continue;
        }
        key = (CFStringRef)keys[i];
        value = (CFDataRef)values[i];

        if (!CFStringGetFileSystemRepresentation(key, name, sizeof(name))) {
            // Skip key
            continue;
        }

        // If this fails, we don't care
        sysctlbyname(name, NULL, NULL, (void *)CFDataGetBytePtr(value), CFDataGetLength(value));
    }

    free(keys);
    free(values);

out:
    return BASErrnoToOSStatus(retval);
}

/* NX Changes */

int CreateMountPathHandler(AuthorizationRef        auth,
                                       const void             *userData,
                                       CFDictionaryRef         request,
                                       CFMutableDictionaryRef  response,
                                       aslclient               asl,
                                       aslmsg                  aslMsg)
{
    int retval = 0;
    
    CFStringRef mountPath = NULL;
    char mountPathC[MAXPATHLEN];
    char realMountPathC[MAXPATHLEN];
    
    CFNumberRef uidNumber;
    uid_t uid;
    CFNumberRef gidNumber;
    gid_t gid;
    
    //assert(auth != NULL);
    assert(request != NULL);
    //assert(response != NULL);

    if (!CFDictionaryGetValueIfPresent(request, CFSTR(kCreateMountPointRequestKeyPath), (const void **)&mountPath)) {
        retval = EINVAL;
        goto out;
    }

    if (!CFStringGetFileSystemRepresentation(mountPath, mountPathC, sizeof(mountPathC))) {
        retval = ENAMETOOLONG;
        goto out;
    }
    
    if (!CFDictionaryGetValueIfPresent(request, CFSTR(kCreateMountPointRequestKeyUID), (const void **)&uidNumber)
        || !CFNumberGetValue(uidNumber, kCFNumberSInt32Type, &uid)) {
        retval = EINVAL;
        goto out;
    }

    if (!CFDictionaryGetValueIfPresent(request, CFSTR(kCreateMountPointRequestKeyGID), (const void **)&gidNumber)
        || !CFNumberGetValue(gidNumber, kCFNumberSInt32Type, &gid)) {
        retval = EINVAL;
        goto out;
    }
    
    //asl_log(asl, aslMsg, ASL_LEVEL_DEBUG, "Creating mount path: %s", mountPathC);
    
    (void)seteuid(uid);
    (void)setegid(gid);
    
    while (true) {
        struct stat sb;

        if (realpath(mountPathC, realMountPathC) != NULL
            && stat(realMountPathC, &sb) == 0) {

            if (S_ISDIR(sb.st_mode)) {
                break;
            } else {
                //asl_log(asl, aslMsg, ASL_LEVEL_DEBUG, "%s: not a directory", realMountPathC);
                errx(EX_USAGE, "%s: not a directory", realMountPathC);
                retval = ENOTDIR;
                goto out_uid;
            }
            
        } else if (errno == ENOENT) {
            bool volumes = strncmp(realMountPathC, "/Volumes/", 9) == 0
                           && strchr(realMountPathC + 9, '/') == NULL;
            
            if (volumes) {
                (void)seteuid(0);
                (void)setegid(0);
            }
            
            if (mkdir(realMountPathC, 0755)) {
                //asl_log(asl, aslMsg, ASL_LEVEL_DEBUG, "%s: %s", realMountPathC, strerror(errno));
                errx(EX_USAGE, "%s: %s", realMountPathC, strerror(errno));
                retval = errno;
                goto out_uid;
            }
            
            if (volumes) {
                (void)chown(realMountPathC, uid, gid);
                (void)seteuid(uid);
                (void)setegid(gid);
            }
            
        } else {
            //asl_log(asl, aslMsg, ASL_LEVEL_DEBUG, "%s: %s", realMountPathC, strerror(errno));
            errx(EX_USAGE, "%s: %s", realMountPathC, strerror(errno));
            retval = errno;
            goto out_uid;
        }
    }

out_uid:
    (void)seteuid(0);
    (void)setegid(0);

out:
    return retval;
}

#pragma mark Privileged Helper Infrastructure

/*
 * This array must be exactly parallel to the kMounterCommandSet array
 * in "MounterCommon.c".
 */
static const BASCommandProc kHelperCommandProcs[] = {
    &GetVersionHandler,
    &LoadKextHandler,
    &UnloadKextHandler,
    &LoadSysctlHandler,
    &CreateMountPathHandler,
    NULL
};

#if 0

int main(int argc, char *argv[])
{
    /*
     * BASHelperToolMain doesn't clean up after itself, so once it returns
     * we must quit.
     */
	return BASHelperToolMain(kMounterCommandSet, kHelperCommandProcs);
}

#endif

/* NX */
