//
//  Mounter.m
//  Mounter
//
//  Created by Benjamin Fleischer on 27.11.12.
//  Copyright (c) 2012-2019 Benjamin Fleischer. All rights reserved.
//

#include <assert.h>
#include <AssertMacros.h>
#include <err.h>
#include <fcntl.h>
#include <fsproperties.h>
#include <getopt.h>
#include <grp.h>
#include <libgen.h>
#include <limits.h>
#include <mach/mach.h>
#include <netinet/in.h>
#include <paths.h>
#include <signal.h>
#include <stdbool.h>
#include <stddef.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/attr.h>
#include <sys/mount.h>
#include <sys/socket.h>
#include <sys/stat.h>
#include <sys/statvfs.h>
#include <sys/sysctl.h>
#include <sys/types.h>
#include <sys/un.h>
#include <sys/utsname.h>
#include <sys/vnode.h>
#include <sysexits.h>
#include <unistd.h>

#include <AppKit/AppKit.h>
#include <CoreFoundation/CoreFoundation.h>
#include <CoreServices/CoreServices.h>

#include "fuse_ioctl.h"
#include "fuse_mount.h"
#include "fuse_param.h"
#include "fuse_version.h"

#include "BetterAuthorizationSampleLib.h"
#include "mntopts.h"
#include "MounterCommon.h"

/* NX Changes */
#define STRINGIFY(s) #s
/* NX */

// Bundle identifer of kernel extension loader
#define MOUNTER_BUNDLE_IDENTIFIER STRINGIFY(MOUNTER_BUNDLE_IDENTIFIER_LITERAL)

// Bundle identifer of helper tool
#define HELPER_BUNDLE_IDENTIFIER  STRINGIFY(HELPER_BUNDLE_IDENTIFIER_LITERAL)

// Version number generated with agvtool
extern const double MounterVersionNumber;

static AuthorizationRef gAuth;

static bool install_only = false;
static bool load_only = false;
static bool quiet_mode = false;
static int signal_fd  = -1;

void showhelp(void);
void showversion(int doexit);

/* NX Changes */
extern int CreateMountPathHandler(AuthorizationRef auth, const void *userData, CFDictionaryRef request, 
    CFMutableDictionaryRef response, aslclient asl, aslmsg aslMsg);
/* NX */

struct mntopt mopts[] = {
    MOPT_STDOPTS,
    MOPT_UPDATE,
    { "allow_other",         0, FUSE_MOPT_ALLOW_OTHER,            1 }, // kused
    { "allow_recursion",     0, FUSE_MOPT_ALLOW_RECURSION,        1 }, // uused
    { "allow_root",          0, FUSE_MOPT_ALLOW_ROOT,             1 }, // kused
    { "auto_cache",          0, FUSE_MOPT_AUTO_CACHE,             1 }, // kused
    { "auto_xattr",          0, FUSE_MOPT_AUTO_XATTR,             1 }, // kused
    { "blocksize=",          0, FUSE_MOPT_BLOCKSIZE,              1 }, // kused
    { "daemon_timeout=",     0, FUSE_MOPT_DAEMON_TIMEOUT,         1 }, // kused
    { "debug",               0, FUSE_MOPT_DEBUG,                  1 }, // kused
    { "default_permissions", 0, FUSE_MOPT_DEFAULT_PERMISSIONS,    1 }, // kused
    { "defer_permissions",   0, FUSE_MOPT_DEFER_PERMISSIONS,      1 }, // kused
    { "direct_io",           0, FUSE_MOPT_DIRECT_IO,              1 }, // kused
    { "excl_create",         0, FUSE_MOPT_EXCL_CREATE,            1 }, // kused
    { "extended_security",   0, FUSE_MOPT_EXTENDED_SECURITY,      1 }, // kused
    { "fsid=" ,              0, FUSE_MOPT_FSID,                   1 }, // kused
    { "fsname=",             0, FUSE_MOPT_FSNAME,                 1 }, // kused
    { "fssubtype=",          0, FUSE_MOPT_FSSUBTYPE,              1 }, // kused
    { "fstypename=",         0, FUSE_MOPT_FSTYPENAME,             1 }, // kused
    { "iosize=",             0, FUSE_MOPT_IOSIZE,                 1 }, // kused
    { "jail_symlinks",       0, FUSE_MOPT_JAIL_SYMLINKS,          1 }, // kused
    { "local",               0, FUSE_MOPT_LOCALVOL,               1 }, // kused
    { "native_xattr",        0, FUSE_MOPT_NATIVE_XATTR,           1 }, // kused
    { "negative_vncache",    0, FUSE_MOPT_NEGATIVE_VNCACHE,       1 }, // kused
    { "sparse",              0, FUSE_MOPT_SPARSE,                 1 }, // kused
    { "slow_statfs",         0, FUSE_MOPT_SLOW_STATFS,            1 }, // kused
    { "use_ino",             0, FUSE_MOPT_USE_INO,                1 },
    { "volname=",            0, FUSE_MOPT_VOLNAME,                1 }, // kused
    
    /* negative ones */
    
    { "appledouble",         1, FUSE_MOPT_NO_APPLEDOUBLE,         1 }, // kused
    { "applexattr",          1, FUSE_MOPT_NO_APPLEXATTR,          1 }, // kused
    { "attrcache",           1, FUSE_MOPT_NO_ATTRCACHE,           1 }, // kused
    { "autonotify",          1, FUSE_MOPT_NO_AUTO_NOTIFY,         1 }, // kused
    { "browse",              1, FUSE_MOPT_NO_BROWSE,              1 }, // kused
    { "localcaches",         1, FUSE_MOPT_NO_LOCALCACHES,         1 }, // kused
    { "readahead",           1, FUSE_MOPT_NO_READAHEAD,           1 }, // kused
    { "synconclose",         1, FUSE_MOPT_NO_SYNCONCLOSE,         1 }, // kused
    { "syncwrites",          1, FUSE_MOPT_NO_SYNCWRITES,          1 }, // kused
    { "ubc",                 1, FUSE_MOPT_NO_UBC,                 1 }, // kused
    { "vncache",             1, FUSE_MOPT_NO_VNCACHE,             1 }, // kused
    
    { NULL }
};

typedef int (* converter_t)(void **target, void *value, void *fallback);

struct mntval {
    uint64_t    mv_mntflag;
    void       *mv_value;
    size_t      mv_len;
    converter_t mv_converter;
    void       *mv_fallback;
    void      **mv_target;
    char       *mv_errstr;
};

static __inline__ int
fuse_to_string(void **target, void *value, void *fallback)
{
    if (!value) {
        // think about what to do if we want to set a NULL value when the
        // fallback value is non-NULL
        value = fallback;
    }
    
    *target = value;
    
    return 0;
}

static __inline__ int
fuse_to_uint32(void **target, void *value, void *fallback)
{
    unsigned long u;
    
    if (!value) {
        *target = fallback;
        return 0;
    }
    
    errno = 0;
    u = strtoul((char *)value, NULL, 10);
    if ((errno == ERANGE) || (errno == EINVAL)) {
        *target = fallback;
        return errno;
    }
    
    *target = (void *)u;
    
    return 0;
}

static __inline__ int
fuse_to_fsid(void **target, void *value, void *fallback)
{
    int ret;
    uint32_t u;
    
    if (!value) {
        *target = fallback;
        return 0;
    }
    
    ret = fuse_to_uint32(target, value, fallback);
    
    if (ret) {
        return ret;
    }
    
    u = *(uint32_t *)target;
    
    if ((u & ~FUSE_MINOR_MASK) || (u == 0)) {
        return EINVAL;
    }
    
    return 0;
}

static uintptr_t blocksize      = FUSE_DEFAULT_BLOCKSIZE;
static uintptr_t daemon_timeout = FUSE_DEFAULT_DAEMON_TIMEOUT;
static uintptr_t fsid           = 0;
static char     *fsname         = NULL;
static uintptr_t fssubtype      = 0;
static char     *fstypename     = NULL;
static uintptr_t iosize         = FUSE_DEFAULT_IOSIZE;
static uint32_t  drandom        = 0;
static char     *volname        = NULL;

struct mntval mvals[] = {
    {
        FUSE_MOPT_BLOCKSIZE,
        NULL,
        0,
        fuse_to_uint32,
        (void *)FUSE_DEFAULT_BLOCKSIZE,
        (void **)&blocksize,
        "invalid value for argument blocksize"
    },
    {
        FUSE_MOPT_DAEMON_TIMEOUT,
        NULL,
        0,
        fuse_to_uint32,
        (void *)FUSE_DEFAULT_DAEMON_TIMEOUT,
        (void **)&daemon_timeout,
        "invalid value for argument daemon_timeout"
    },
    {
        FUSE_MOPT_FSID,
        NULL,
        0,
        fuse_to_fsid,
        0,
        (void **)&fsid,
        "invalid value for argument fsid (must be 0 < fsid < 0xFFFFFF)"
    },
    {
        FUSE_MOPT_FSNAME,
        NULL,
        0,
        fuse_to_string,
        NULL,
        (void **)&fsname,
        "invalid value for argument fsname"
    },
    {
        FUSE_MOPT_IOSIZE,
        NULL,
        0,
        fuse_to_uint32,
        (void *)FUSE_DEFAULT_IOSIZE,
        (void **)&iosize,
        "invalid value for argument iosize"
    },
    {
        FUSE_MOPT_FSSUBTYPE,
        NULL,
        0,
        fuse_to_uint32,
        NULL,
        (void **)&fssubtype,
        "invalid value for argument fssubtype"
    },
    {
        FUSE_MOPT_FSTYPENAME,
        NULL,
        0,
        fuse_to_string,
        NULL,
        (void **)&fstypename,
        "invalid value for argument fstypename"
    },
    {
        FUSE_MOPT_VOLNAME,
        NULL,
        0,
        fuse_to_string,
        NULL,
        (void **)&volname,
        "invalid value for argument volname"
    },
    {
        0, NULL, 0, NULL, (void *)NULL, (void **)NULL, (char *)NULL
    },
};

static void
fuse_process_mvals(void)
{
    int ret;
    struct mntval *mv;
    
    for (mv = mvals; mv->mv_mntflag; mv++) {
        ret = mv->mv_converter(mv->mv_target, mv->mv_value, mv->mv_fallback);
        if (ret) {
            errx(EX_USAGE, "%s", mv->mv_errstr);
        }
    }
}

/* osxfuse notifications */

enum osxfuse_notification {
    NOTIFICATION_OS_IS_TOO_NEW,
    NOTIFICATION_OS_IS_TOO_OLD,
    NOTIFICATION_VERSION_MISMATCH,
    NOTIFICATION_SYSTEM_POLICY,
    NOTIFICATION_MOUNT
};
typedef enum osxfuse_notification osxfuse_notification_t;

const char * const osxfuse_notification_names[] = {
    "k" OSXFUSE_DISPLAY_NAME "OSIsTooNew",      // NOTIFICATION_OS_IS_TOO_NEW
    "k" OSXFUSE_DISPLAY_NAME "OSIsTooOld",      // NOTIFICATION_OS_IS_TOO_OLD
    "k" OSXFUSE_DISPLAY_NAME "VersionMismatch", // NOTIFICATION_VERSION_MISMATCH
    "k" OSXFUSE_DISPLAY_NAME "SystemPolicy",    // NOTIFICATION_SYSTEM_POLICY
    "k" OSXFUSE_DISPLAY_NAME "Mount"            // NOTIFICATION_MOUNT
};

const char * const osxfuse_notification_object = OSXFUSE_IDENTIFIER;

/* User info keys */

#define kFUSEDevicePathKey "kFUSEDevicePath"
#define kFUSEMountPathKey  "kFUSEMountPath"

static void
post_notification(const osxfuse_notification_t  notification,
                  const char                   *dict[][2],
                  const int                     dict_count)
{
    CFNotificationCenterRef notification_center =
    CFNotificationCenterGetDistributedCenter();
    
    CFStringRef            name      = NULL;
    CFStringRef            object    = NULL;
    CFMutableDictionaryRef user_info = NULL;
    
    name   = CFStringCreateWithCString(kCFAllocatorDefault,
                                       osxfuse_notification_names[notification],
                                       kCFStringEncodingUTF8);
    object = CFStringCreateWithCString(kCFAllocatorDefault,
                                       osxfuse_notification_object,
                                       kCFStringEncodingUTF8);
    
    if (!name || !object) goto out;
    if (dict_count == 0)  goto post;
    
    user_info = CFDictionaryCreateMutable(kCFAllocatorDefault, dict_count,
                                          &kCFCopyStringDictionaryKeyCallBacks,
                                          &kCFTypeDictionaryValueCallBacks);
    
    CFStringRef key;
    CFStringRef value;
    int         i;
    for (i = 0; i < dict_count; i++) {
        key   = CFStringCreateWithCString(kCFAllocatorDefault, dict[i][0],
                                          kCFStringEncodingUTF8);
        value = CFStringCreateWithCString(kCFAllocatorDefault, dict[i][1],
                                          kCFStringEncodingUTF8);
        
        if (!key || !value) {
            if (key)   CFRelease(key);
            if (value) CFRelease(value);
            goto out;
        }
        
        CFDictionarySetValue(user_info, key, value);
        CFRelease(key); key = NULL;
        CFRelease(value); value = NULL;
    }
    
post:
    CFNotificationCenterPostNotification(notification_center, name, object,
                                         user_info, false);
    out:
    if (name)      CFRelease(name);
    if (object)    CFRelease(object);
    if (user_info) CFRelease(user_info);
}

static long
fuse_os_version_major_np(void)
{
    int ret = 0;
    long major = 0;
    char *c = NULL;
    struct utsname u;
    size_t oldlen;
    
    oldlen = sizeof(u.release);
    
    ret = sysctlbyname("kern.osrelease", u.release, &oldlen, NULL, 0);
    if (ret != 0) {
        return -1;
    }
    
    c = strchr(u.release, '.');
    if (c == NULL) {
        return -1;
    }
    
    *c = '\0';
    
    errno = 0;
    major = strtol(u.release, NULL, 10);
    if ((errno == EINVAL) || (errno == ERANGE)) {
        return -1;
    }
    
    return major;
}

static inline OSStatus ExecuteRequest(CFStringRef command, CFMutableDictionaryRef request, CFDictionaryRef *response)
{
    OSStatus err = noErr;

    Boolean releaseRequest = false;
    CFDictionaryRef responseFallback = NULL;

    CFNumberRef version;

    if (!request) {
        request = CFDictionaryCreateMutable(kCFAllocatorDefault, 2, &kCFTypeDictionaryKeyCallBacks, &kCFTypeDictionaryValueCallBacks);
        releaseRequest = true;
    }
    if (!response) {
        response = &responseFallback;
    }

    // Command name
    CFDictionarySetValue(request, CFSTR(kBASCommandKey), command);

    // Command version
    version = CFNumberCreate(kCFAllocatorDefault, kCFNumberDoubleType, &MounterVersionNumber);
    CFDictionarySetValue(request, CFSTR(kBASCommandVersionKey), version);
    CFRelease(version);

    // Execute request
    err = BASExecuteRequestInHelperTool(gAuth,
                                        kMounterCommandSet,
                                        CFSTR(HELPER_BUNDLE_IDENTIFIER),
                                        request,
                                        response);
    if (err == noErr) {
        err = BASGetErrorFromResponse(*response);
    }

    if (releaseRequest) {
        CFRelease(request);
    }
    if (responseFallback) {
        CFRelease(responseFallback);
    }

    return err;
}

static int InstallHelperTool(void)
{
    OSStatus err = 0;
    BASFailCode failCode = kBASFailUnknown;

    CFDictionaryRef response = NULL;

    // Execute request
    err = ExecuteRequest(CFSTR(kGetVersionCommand), NULL, &response);

    if (err == noErr) {
        CFNumberRef version;

        assert(response != NULL);

        version = CFDictionaryGetValue(response, CFSTR(kGetVersionResponseKeyVersion));
        {
            double v;

            CFNumberGetValue(version, kCFNumberDoubleType, &v);
            if (v < MounterVersionNumber) {
                failCode = kBASFailNeedsUpdate;
            }
        }
    }

    if ((err != noErr && err != userCanceledErr) || failCode == kBASFailNeedsUpdate) {
        // Install helper tool

        CFStringRef helperToolName = CFSTR(HELPER_BUNDLE_IDENTIFIER);
        CFBundleRef mainBundle = NULL;
        CFStringRef bundleId = NULL;

        // Get bundle ID
        mainBundle = CFBundleGetMainBundle();
        if (!mainBundle) {
            goto out;
        }
        bundleId = CFBundleGetIdentifier(mainBundle);
        if (!bundleId) {
            goto out;
        }

        if (failCode == kBASFailUnknown) {
            // Diagnose failure
            failCode = BASDiagnoseFailure(gAuth, helperToolName);
        }
        err = BASFixFailure(gAuth, bundleId, helperToolName, failCode);
    }

out:
    if (response) {
        CFRelease(response);
    }

    return err;
}

static CFStringRef CreateStringByResolvingPath(CFStringRef path)
{
    char pathC[PATH_MAX];
    char resolvedPathC[PATH_MAX];

    if (!CFStringGetCString(path, pathC, PATH_MAX, kCFStringEncodingUTF8)) {
        return NULL;
    }

    if (realpath(pathC, resolvedPathC) == NULL) {
        return NULL;
    }

     return CFStringCreateWithCString(kCFAllocatorDefault, resolvedPathC, kCFStringEncodingUTF8);
}

static OSStatus LoadKext(CFURLRef kextUrl)
{
    OSStatus retval = 0;
    CFStringRef absoluteKextPath = NULL;
    CFStringRef resolvedKextPath = NULL;
    CFMutableDictionaryRef request = NULL;

    // Get absolute path
    CFURLRef absoluteKextUrl = CFURLCopyAbsoluteURL(kextUrl);
    absoluteKextPath = CFURLCopyFileSystemPath(absoluteKextUrl, kCFURLPOSIXPathStyle);
    CFRelease(absoluteKextUrl);

    // Resolve path
    resolvedKextPath = CreateStringByResolvingPath(absoluteKextPath);
    if (!resolvedKextPath) {
        retval = BASErrnoToOSStatus(ENOENT);
        goto out;
    }

    // Make request
    request = CFDictionaryCreateMutable(kCFAllocatorDefault, 0, &kCFTypeDictionaryKeyCallBacks, &kCFTypeDictionaryValueCallBacks);
    CFDictionarySetValue(request, CFSTR(kLoadKextRequestKeyPath), resolvedKextPath);

    // Execute request
    retval = ExecuteRequest(CFSTR(kLoadKextCommand), request, NULL);

    CFRelease(resolvedKextPath);
    CFRelease(request);

out:
    CFRelease(absoluteKextPath);

    return retval;
}

static OSStatus UnloadKext(void)
{
    // Execute request
    return ExecuteRequest(CFSTR(kUnloadKextCommand), NULL, NULL);
}

static OSStatus SetSysctls(CFDictionaryRef sysctls)
{
    OSStatus retval = 0;
    CFMutableDictionaryRef request = NULL;

    if (!sysctls) {
        retval = BASErrnoToOSStatus(EINVAL);
        goto out;
    }

    // Make request
    request = CFDictionaryCreateMutable(kCFAllocatorDefault, 0, &kCFTypeDictionaryKeyCallBacks, &kCFTypeDictionaryValueCallBacks);
    CFDictionarySetValue(request, CFSTR(kSetSysctlsRequestKeySysctls), sysctls);

    // Execute request
    retval = ExecuteRequest(CFSTR(kSetSysctlsCommand), request, NULL);

    CFRelease(request);

out:
    return retval;
}

static int LoadFileSystemKext(void)
{
    int retval = 0;
    int systemVersionMajor;
    int systemVersionMinor;

    struct vfsconf vfc = { 0 };

    char version[MAXHOSTNAMELEN + 1] = { 0 };
    size_t version_len = MAXHOSTNAMELEN;
    size_t version_len_desired = 0;

    int32_t tunables_initialized = 0;
    size_t tunables_initialized_len = sizeof(tunables_initialized);

    CFBundleRef mainBundle = NULL;
    CFURLRef bundleURL = NULL;
    CFStringRef kextDirectoryName = NULL;
    CFURLRef kextDirectoryUrl = NULL;
    CFURLRef kextUrl;

    if (!GetSystemVersion(&systemVersionMajor, &systemVersionMinor, NULL)) {
        /*
         * In order to load the correct kernel extension we need to determine
         * the version of macOS. Since we cannot figure out which version is
         * running return ENOENT.
         */
        retval = ENOENT;
        goto out;
    }

    retval = getvfsbyname(OSXFUSE_NAME, &vfc);
    if (retval) {
        // File system kernel extension is not already loaded.
        goto load_kext;
    }

    /*
     * Note: Some version of our kernel extension is already loaded. We need to
     * check it out.
     */

    retval = sysctlbyname(OSXFUSE_SYSCTL_VERSION_NUMBER, version, &version_len, NULL, (size_t)0);
    if (retval) {
        retval = -1;
        goto out;
    }

    // sysctlbyname includes the trailing '\0' in version_len
    version_len_desired = strlen(OSXFUSE_VERSION) + 1;

    if (version_len == version_len_desired && !strncmp(OSXFUSE_VERSION, version, version_len)) {
        // Currently loaded kernel extension is good
        retval = 0;
        goto initialize_kext;
    }

    /*
     * Version mismatch. We need to unload the currently loaded version of the
     * kernel extension.
     */
    retval = UnloadKext();

    if (retval) {
        // Unloading failed
        retval = EBUSY;
        goto out;
    }

    // Unloading succeeded. Now load the on-disk version.

load_kext:
    // Locate kernel extension
    mainBundle = CFBundleGetMainBundle();
    if (!mainBundle) {
        retval = ENOENT;
        goto out;
    }
    bundleURL = CFBundleCopyBundleURL(mainBundle);
    if (!bundleURL) {
        retval = ENOENT;
        goto out;
    }
    if (systemVersionMajor > 10) {
        kextDirectoryName = CFStringCreateWithFormat(kCFAllocatorDefault, NULL, CFSTR("Contents/Extensions/%d"), systemVersionMajor);
    } else {
        // Use old version numbering scheme
        kextDirectoryName = CFStringCreateWithFormat(kCFAllocatorDefault, NULL, CFSTR("Contents/Extensions/%d.%d"), systemVersionMajor, systemVersionMinor);
    }
    kextDirectoryUrl = CFURLCreateCopyAppendingPathComponent(kCFAllocatorDefault, bundleURL, kextDirectoryName, true);
    kextUrl = CFURLCreateCopyAppendingPathComponent(kCFAllocatorDefault, kextDirectoryUrl, CFSTR(OSXFUSE_KEXT_NAME), true);

    // Load kernel extension
    retval = LoadKext(kextUrl);

    CFRelease(bundleURL);
    CFRelease(kextDirectoryName);
    CFRelease(kextDirectoryUrl);
    CFRelease(kextUrl);

    if (@available(macOS 11, *)) {
        /*
         * KextManagerLoadKextWithIdentifier() is broken on macOS 11 and might
         * return an error even if the kernel extension has been loaded
         * successfully. We need to double-check if the kernel extension has
         * been loaded.
         */
        if (retval && !getvfsbyname(OSXFUSE_NAME, &vfc) && !sysctlbyname(OSXFUSE_SYSCTL_VERSION_NUMBER, NULL, NULL, NULL, (size_t)0)) {
            retval = 0;
        }
    }

    if (retval) {
        retval = BASOSStatusToErrno(retval);
        goto out;
    }

initialize_kext:
    (void)sysctlbyname(OSXFUSE_SYSCTL_CONTROL_INIT, &tunables_initialized, &tunables_initialized_len, NULL, (size_t)0);

    if (!tunables_initialized) {
        CFMutableDictionaryRef sysctls;
        struct group *adminGroup;

        sysctls = CFDictionaryCreateMutable(kCFAllocatorDefault, 0, &kCFTypeDictionaryKeyCallBacks, &kCFTypeDictionaryValueCallBacks);

        adminGroup = getgrnam(MACOSX_ADMIN_GROUP_NAME);
        if (adminGroup) {
            CFDataRef value = CFDataCreate(kCFAllocatorDefault, (const UInt8 *)&adminGroup->gr_gid, sizeof(adminGroup->gr_gid));
            CFDictionaryAddValue(sysctls, CFSTR(OSXFUSE_SYSCTL_TUNABLES_ADMIN), value);
            CFRelease(value);
        }

        tunables_initialized = 1;
        {
            CFDataRef value = CFDataCreate(kCFAllocatorDefault, (const UInt8 *)&tunables_initialized, sizeof(tunables_initialized));
            CFDictionaryAddValue(sysctls, CFSTR(OSXFUSE_SYSCTL_CONTROL_INIT), value);
            CFRelease(value);
        }

        retval = SetSysctls(sysctls);
        if (retval) {
            retval = ENXIO;
            goto out;
        }
    }

out:
    if (retval) {
        if (retval == EINVAL) {
            if (!quiet_mode) {
                CFUserNotificationDisplayNotice(
                    (CFTimeInterval)0,
                    kCFUserNotificationCautionAlertLevel,
                    (CFURLRef)NULL,
                    (CFURLRef)NULL,
                    (CFURLRef)NULL,
                    CFSTR("Unsupported macOS Version"),
                    CFSTR("The installed version of " OSXFUSE_MOUNTER_DISPLAY_NAME " is too new for the operating system. Please downgrade your " OSXFUSE_MOUNTER_DISPLAY_NAME " installation to one that is compatible with the currently running version of macOS."),
                    CFSTR("OK"));
            }
            post_notification(NOTIFICATION_OS_IS_TOO_OLD, NULL, 0);
        }
        if (retval == ENOENT) {
            if (!quiet_mode) {
                CFUserNotificationDisplayNotice(
                    (CFTimeInterval)0,
                    kCFUserNotificationCautionAlertLevel,
                    (CFURLRef)NULL,
                    (CFURLRef)NULL,
                    (CFURLRef)NULL,
                    CFSTR("Unsupported macOS Version"),
                    CFSTR("The installed version of " OSXFUSE_MOUNTER_DISPLAY_NAME " is too old for the operating system. Please upgrade your " OSXFUSE_MOUNTER_DISPLAY_NAME " installation to one that is compatible with the currently running version of macOS."),
                    CFSTR("OK"));
            }
            post_notification(NOTIFICATION_OS_IS_TOO_NEW, NULL, 0);
        } else if (retval == EBUSY) {
            if (!quiet_mode) {
                CFUserNotificationDisplayNotice(
                    (CFTimeInterval)0,
                    kCFUserNotificationCautionAlertLevel,
                    (CFURLRef)NULL,
                    (CFURLRef)NULL,
                    (CFURLRef)NULL,
                    CFSTR("Version Mismatch"),
                    CFSTR(OSXFUSE_MOUNTER_DISPLAY_NAME " has been updated but an incompatible or old version of the system extension is already loaded. It failed to unload, possibly because a " OSXFUSE_MOUNTER_DISPLAY_NAME " volume is currently mounted.\n\nPlease eject all " OSXFUSE_MOUNTER_DISPLAY_NAME " volumes and try again, or simply restart the system for changes to take effect."),
                    CFSTR("OK"));
            }
            post_notification(NOTIFICATION_VERSION_MISMATCH, NULL, 0);
        } else if (retval == EPERM) {
            if (!quiet_mode) {
                CFOptionFlags response_flags = 0;
                CFUserNotificationDisplayAlert(
                    (CFTimeInterval)0,
                    kCFUserNotificationCautionAlertLevel,
                    (CFURLRef)NULL,
                    (CFURLRef)NULL,
                    (CFURLRef)NULL,
                    CFSTR("System Extension Blocked"),
                    CFSTR("The system extension required for mounting " OSXFUSE_MOUNTER_DISPLAY_NAME " volumes could not be loaded.\n\nPlease open the Security & Privacy System Preferences pane, go to the General preferences and allow loading system software from developer \"" OSXFUSE_CODESIGN_SUBJECT_CN "\". A restart might be required before the system extension can be loaded.\n\nThen try mounting the volume again."),
                    CFSTR("Open System Preferences"),
                    CFSTR("Cancel"),
                    NULL,
                    &response_flags);
                
                if (response_flags == kCFUserNotificationDefaultResponse) {
                    NSURL *systemPreferencesURL = [NSURL URLWithString:@"x-apple.systempreferences:com.apple.preference.security?General"];
                    [[NSWorkspace sharedWorkspace] openURL:systemPreferencesURL];
                }
            }
            post_notification(NOTIFICATION_SYSTEM_POLICY, NULL, 0);
        }
        errx(EX_UNAVAILABLE, "the " OSXFUSE_DISPLAY_NAME " file system is not available (%d)", retval);
    }

    return retval;
}

static int
check_kext_status(void)
{
    int    result = -1;
    char   version[MAXHOSTNAMELEN + 1] = { 0 };
    size_t version_len = MAXHOSTNAMELEN;
    size_t version_len_desired = 0;
    struct vfsconf vfc = { 0 };
    
    result = getvfsbyname(OSXFUSE_NAME, &vfc);
    if (result) { /* osxfuse is not already loaded */
        return ESRCH;
    }
    
    /* Some version of osxfuse is already loaded. Let's check it out */
    
    result = sysctlbyname(OSXFUSE_SYSCTL_VERSION_NUMBER, version,
                          &version_len, (void *)NULL, (size_t)0);
    if (result) {
        return result;
    }
    
    /* sysctlbyname() includes the trailing '\0' in version_len */
    version_len_desired = strlen(OSXFUSE_VERSION) + 1;
    
    if ((version_len != version_len_desired) ||
        strncmp(OSXFUSE_VERSION, version, version_len)) {
        return EINVAL;
    }
    
    /* What's currently loaded is good */
    
    return 0;
}

static int
send_fd(int sock_fd, int fd)
{
    ssize_t retval;
    
    struct iovec vec;
    char sendchar = 0;
    
    struct msghdr msg;
    char cmsgbuf[CMSG_SPACE(sizeof(fd))];
    struct cmsghdr *cmsgp;
    
    vec.iov_base = &sendchar;
    vec.iov_len = sizeof(sendchar);
    
    memset(cmsgbuf, 0, sizeof(cmsgbuf));
    
    msg.msg_name = NULL;
    msg.msg_namelen = 0;
    msg.msg_iov = &vec;
    msg.msg_iovlen = 1;
    msg.msg_control = cmsgbuf;
    msg.msg_controllen = (socklen_t)sizeof(cmsgbuf);
    msg.msg_flags = 0;
    
    cmsgp = CMSG_FIRSTHDR(&msg);
    cmsgp->cmsg_len = (socklen_t)CMSG_LEN(sizeof(fd));
    cmsgp->cmsg_level = SOL_SOCKET;
    cmsgp->cmsg_type = SCM_RIGHTS;
    
    memcpy(CMSG_DATA(cmsgp), &fd, sizeof(fd));
    
    msg.msg_controllen = cmsgp->cmsg_len;
    
    while ((retval = sendmsg(sock_fd, &msg, 0)) == -1 && errno == EINTR);
    if (retval != 1) {
        perror("sending file descriptor");
        return -1;
    }
    
    return 0;
}

static int
send_mount_status(int sock_fd, int32_t status) {
    ssize_t retval = 0;
    
    while ((retval = send(sock_fd, &status, sizeof(status), 0)) == -1 &&
           errno == EINTR);
    if (retval != sizeof(status)) {
        perror("sending mount status");
        return -1;
    }
    
    return 0;
}

static void
signal_idx_atexit_handler(void)
{
    if (signal_fd != -1) {
        
        (void)ioctl(signal_fd, FUSEDEVIOCSETDAEMONDEAD, &signal_fd);
        
        /*
         * Originally, I did kill_fs from here.
         *
         * int32_t kill_fs_old = 0;
         * int32_t kill_fs_new = signal_idx;
         * size_t oldlen = sizeof(kill_fs_old);
         * size_t newlen = sizeof(kill_fs_new);
         *
         * (void)sysctlbyname("vfs.generic.osxfuse.control.kill_fs",
         *                    (void *)&kill_fs_old, &oldlen,
         *                    (void *)&kill_fs_new, newlen);
         */
    }
}

void
showhelp()
{
    if (!getenv("_FUSE_CALL_BY_LIB")) {
        showversion(0);
        fprintf(stderr, "\nThis program is not meant to be called directly. The "
                OSXFUSE_DISPLAY_NAME " library calls it.\n");
    }
    fprintf(stderr,
            "\nAvailable mount options:\n"
            "    -o allow_other         allow access to others besides the user who mounted\n"
            "                           the file system\n"
            "    -o allow_recursion     allow a mount point that itself resides on a " OSXFUSE_DISPLAY_NAME "\n"
            "                           volume (by default, such mounting is disallowed)\n"
            "    -o allow_root          allow access to root (can't be used with allow_other)\n"
            "    -o auto_xattr          handle extended attributes entirely through ._ files\n"
            "    -o blocksize=<size>    specify block size in bytes of \"storage\"\n"
            "    -o daemon_timeout=<s>  timeout in seconds for kernel calls to daemon\n"
            "    -o debug               turn on debug information printing\n"
            "    -o default_permissions let the kernel handle permission checks locally\n"
            "    -o defer_permissions   defer permission checks to file operations themselves\n"
            "    -o direct_io           use alternative (direct) path for kernel-user I/O\n"
            "    -o extended_security   turn on macOS extended security (ACLs)\n"
            "    -o fsid=<fsid>         set the second 32-bit component of the fsid\n"
            "    -o fsname=<name>       set the file system's name\n"
            "    -o fssubtype=<num>     set the file system's fssubtype identifier\n"
            "    -o fstypename=<name>   set the file system's type name\n"
            "    -o iosize=<size>       specify maximum I/O size in bytes\n"
            "    -o jail_symlinks       contain symbolic links within the mount\n"
            "    -o local               mark the volume as \"local\" (default is \"nonlocal\")\n"
            "    -o negative_vncache    enable vnode name caching of non-existent objects\n"
            "    -o sparse              enable support for sparse files\n"
            "    -o volname=<name>      set the file system's volume name\n"
            "\nAvailable negative mount options:\n"
            "    -o noalerts            disable all graphical alerts (if any) in " OSXFUSE_DISPLAY_NAME " Core\n"
            "    -o noappledouble       ignore Apple Double (._) and .DS_Store files entirely\n"
            "    -o noapplexattr        ignore all \"com.apple.*\" extended attributes\n"
            "    -o nobrowse            mark the volume as non-browsable by the Finder\n"
            "    -o nolocalcaches       meta option equivalent to noreadahead,noubc,novncache\n"
            "    -o noreadahead         disable I/O read-ahead behavior for this file system\n"
            "    -o nosynconclose       disable sync-on-close behavior (enabled by default)\n"
            "    -o nosyncwrites        disable synchronous-writes behavior (dangerous)\n"
            "    -o noubc               disable the unified buffer cache for this file system\n"
            "    -o novncache           disable the vnode name cache for this file system\n");
    exit(EX_USAGE);
}

void
showversion(int doexit)
{
    fprintf(stderr, "%s mount version %s\n", OSXFUSE_DISPLAY_NAME,
            OSXFUSE_VERSION);
    if (doexit) {
        exit(EX_USAGE);
    }
}

static int
ParseArguments(int argc, char *argv[], int *mntflagsptr, uint64_t *altflagsptr,
               char **mntpathptr, int *cfdptr, int *cversptr)
{
    struct mntopt *mo;
    struct mntval *mv;
    char *commfd;
    char *commvers;
    int cfd;
    int cvers;
    char *endptr;
 
    *mntflagsptr = 0;
    *altflagsptr = 0ULL;
    
    while (true) {
        static struct option long_options[] = {
            { "install", no_argument, NULL, 'i' },
            { "help",    no_argument, NULL, 'h' },
            { "load",    no_argument, NULL, 'l' },
            { "version", no_argument, NULL, 'v' },
            { NULL, 0, NULL, 0 }
        };
        
        int c = getopt_long(argc, argv, "ihlo:qv", long_options, NULL);
        if (c == -1) {
            break;
        }
        
        switch (c) {
            case 'i':
                install_only = true;
                break;
            
            case 'l':
                load_only = true;
                break;
            
            case 'o':
                getmntopts(optarg, mopts, mntflagsptr, altflagsptr);
                for (mv = mvals; mv->mv_mntflag; ++mv) {
                    if (!(*altflagsptr & mv->mv_mntflag)) {
                        continue;
                    }
                    for (mo = mopts; mo->m_option; ++mo) {
                        char *p, *q;
                        if (!mo->m_altloc || mo->m_flag != mv->mv_mntflag) {
                            continue;
                        }
                        p = strstr(optarg, mo->m_option);
                        if (p) {
                            p += strlen(mo->m_option);
                            q = p;
                            while (*q != '\0' && *q != ',') {
                                q++;
                            }
                            mv->mv_len = q - p + 1;
                            mv->mv_value = malloc(mv->mv_len);
                            memcpy(mv->mv_value, p, mv->mv_len - 1);
                            ((char *)mv->mv_value)[mv->mv_len - 1] = '\0';
                            break;
                        }
                    }
                }
                break;
            
            case 'q':
                quiet_mode = true;
                break;
                
            case 'v':
                showversion(true);
                break;
                
            case 'h':
            case '?':
            default:
                showhelp();
                break;
        }
    }
    
    argc -= optind;
    argv += optind;
    
    if (install_only || load_only) {
        return 0;
    }
    
    if (argc > 0) {
        *mntpathptr = argv[0];
    } else {
        fprintf(stderr, "missing mount point\n");
        return -1;
    }

    commfd = getenv("_FUSE_COMMFD");
    if (commfd == NULL) {
        fprintf(stderr, "new style mounting requires commfd\n");
        return -1;
    } else {
        errno = 0;
        cfd = (int)strtol(commfd, &endptr, 10);
        if (endptr == commfd || errno == EINVAL || errno == ERANGE || cfd < 0) {
            fprintf(stderr, "invalid commfd\n");
            return -1;
        }
    }
    
    commvers = getenv("_FUSE_COMMVERS");
    if (commvers == NULL) {
        // Fall back to protocol version 1
        cvers = 1;
    } else {
        errno = 0;
        cvers = (int)strtol(commvers, &endptr, 10);
        if (endptr == commvers || errno == EINVAL || errno == ERANGE || cvers < 1 || cvers > 2) {
            fprintf(stderr, "invalid commvers\n");
            return -1;
        }
    }
    
    *cfdptr = cfd;
    *cversptr = cvers;
    return 0;
}

static int
OpenDevice(int *fdptr, int *dindexptr, int *rdevptr)
{
    int fd = -1;
    int32_t dindex = -1;
    uint32_t rdev;
    
    char *fdnam = NULL;
    char *dev = NULL;
    char devpath[MAXPATHLEN];
    struct stat sb;
    
    fdnam = getenv("FUSE_DEV_FD");
    dev = getenv("FUSE_DEV_NAME");
    
    if (fdnam) {
        char *endptr;
        fd = (int)strtol(fdnam, &endptr, 10);
        if (endptr == fdnam || errno == EINVAL || errno == ERANGE || fd < 0) {
            fprintf(stderr, "invalid value given in FUSE_DEV_FD\n");
            return -1;
        }
        
    } else if (dev) {
        fd = open(dev, O_RDWR);
        if (fd < 0) {
            fprintf(stderr, "failed to open device\n");
            return -1;
        }
        
    } else {
        for (int i = 0; i < OSXFUSE_NDEVICES; i++) {
            snprintf(devpath, MAXPATHLEN - 1,
                     _PATH_DEV OSXFUSE_DEVICE_BASENAME "%d", i);
            fd = open(devpath, O_RDWR);
            if (fd >= 0) {
                dindex = i;
                break;
            }
        }
        if (dindex == -1) {
            fprintf(stderr, "failed to open device\n");
            return -1;
        }
    }
    
    if (fstat(fd, &sb) == -1) {
        fprintf(stderr, "fstat failed for " OSXFUSE_DISPLAY_NAME " device file "
                "descriptor\n");
        goto err_fd;
    }
    rdev = sb.st_rdev;
    
    if (dindex < 0) {
        char  ndev[MAXPATHLEN];
        char *ndevbas;
        char *ndevindex;
        char *endptr;
        
        (void)strlcpy(ndev, _PATH_DEV, sizeof(ndev));
        ndevbas = ndev + strlen(_PATH_DEV);
        devname_r(rdev, S_IFCHR, ndevbas, (int)(sizeof(ndev) - strlen(_PATH_DEV)));
        
        if (strncmp(ndevbas, OSXFUSE_DEVICE_BASENAME,
                    strlen(OSXFUSE_DEVICE_BASENAME))) {
            fprintf(stderr, "mounting inappropriate device");
            goto err_fd;
        }
        
        ndevindex = ndevbas + strlen(OSXFUSE_DEVICE_BASENAME);
        dindex = (int)strtol(ndevindex, &endptr, 10);
        if (endptr == ndevindex || errno == EINVAL || errno == ERANGE
            || dindex < 0 || dindex > OSXFUSE_NDEVICES) {
            fprintf(stderr, "invalid " OSXFUSE_DISPLAY_NAME " device unit "
                    "(%d)\n", dindex);
            goto err_fd;
        }
    }
    
    *fdptr = fd;
    *dindexptr = dindex;
    *rdevptr = rdev;
    return 0;

err_fd:
    if (fd >= 0) {
        (void)ioctl(fd, FUSEDEVIOCSETDAEMONDEAD, &fd);
    }
    
    return -1;
}

static int
CreateMountPath(char *mntpath)
{
    int retval = 0;
    
    CFMutableDictionaryRef request = NULL;
    CFStringRef mountPath;
    uid_t uid;
    CFNumberRef uidNumber;
    gid_t gid;
    CFNumberRef gidNumber;

    mountPath = CFStringCreateWithCString(kCFAllocatorDefault, mntpath, kCFStringEncodingUTF8);
    
    // Make request
    request = CFDictionaryCreateMutable(kCFAllocatorDefault, 0, &kCFTypeDictionaryKeyCallBacks, &kCFTypeDictionaryValueCallBacks);
    CFDictionarySetValue(request, CFSTR(kCreateMountPointRequestKeyPath), mountPath);
    
    uid = getuid();
    uidNumber = CFNumberCreate(kCFAllocatorDefault, kCFNumberSInt32Type, &uid);
    CFDictionarySetValue(request, CFSTR(kCreateMountPointRequestKeyUID), uidNumber);
    
    gid = getgid();
    gidNumber = CFNumberCreate(kCFAllocatorDefault, kCFNumberSInt32Type, &gid);
    CFDictionarySetValue(request, CFSTR(kCreateMountPointRequestKeyGID), gidNumber);
    
    // Execute request
    /* NX Changes */
    //retval = ExecuteRequest(CFSTR(kCreateMountPathCommand), request, NULL);
    retval = CreateMountPathHandler(NULL, NULL, request, NULL, NULL, NULL);
    /* NX */
    
    CFRelease(mountPath);
    CFRelease(uidNumber);
    CFRelease(gidNumber);
    
    CFRelease(request);
    
    return retval;
}

static int
MountFileSystem(int fd, int dindex, fuse_mount_args *argsptr, int mntflags)
{
    int result;
    struct statfs statfsb;
 
    fuse_process_mvals();
    
    if (statfs(argsptr->mntpath, &statfsb)) {
        errx(EX_OSFILE, "cannot stat the mount point %s", argsptr->mntpath);
    }
    
    if (((strlen(statfsb.f_fstypename) == strlen(OSXFUSE_NAME)) &&
         (strcmp(statfsb.f_fstypename, OSXFUSE_NAME) == 0)) ||
        ((strlen(OSXFUSE_TYPE_NAME_PREFIX) > 0) &&
         (strncmp(statfsb.f_fstypename, OSXFUSE_TYPE_NAME_PREFIX,
                  strlen(OSXFUSE_TYPE_NAME_PREFIX)) == 0))) {
            if (!(argsptr->altflags & FUSE_MOPT_ALLOW_RECURSION)) {
                fprintf(stderr, "mount point %s is itself on a "
                     OSXFUSE_DISPLAY_NAME " volume\n", argsptr->mntpath);
                return -1;
            }
        }
    
    /* allow_root and allow_other checks are done in the kernel. */
    
    if (argsptr->altflags & FUSE_MOPT_NO_LOCALCACHES) {
        argsptr->altflags |= FUSE_MOPT_NO_ATTRCACHE;
        argsptr->altflags |= FUSE_MOPT_NO_READAHEAD;
        argsptr->altflags |= FUSE_MOPT_NO_UBC;
        argsptr->altflags |= FUSE_MOPT_NO_VNCACHE;
    }
    
    if ((argsptr->altflags & FUSE_MOPT_NEGATIVE_VNCACHE) &&
        (argsptr->altflags & FUSE_MOPT_NO_VNCACHE)) {
        fprintf(stderr, "'negative_vncache' can't be used with 'novncache'\n");
        return -1;
    }
    
    /*
     * 'nosyncwrites' must not appear with either 'noubc' or 'noreadahead'.
     */
    if ((argsptr->altflags & FUSE_MOPT_NO_SYNCWRITES) &&
        (argsptr->altflags & (FUSE_MOPT_NO_UBC | FUSE_MOPT_NO_READAHEAD))) {
        fprintf(stderr,
             "disabling local caching can't be used with 'nosyncwrites'\n");
        return -1;
    }
    
    /*
     * 'nosynconclose' only allowed if 'nosyncwrites' is also there.
     */
    if ((argsptr->altflags & FUSE_MOPT_NO_SYNCONCLOSE) &&
        !(argsptr->altflags & FUSE_MOPT_NO_SYNCWRITES)) {
        fprintf(stderr, "the 'nosynconclose' option requires 'nosyncwrites'\n");
        return -1;
    }
    
    if ((argsptr->altflags & FUSE_MOPT_DEFAULT_PERMISSIONS) &&
        (argsptr->altflags & FUSE_MOPT_DEFER_PERMISSIONS)) {
        fprintf(stderr,
             "'default_permissions' can't be used with 'defer_permissions'\n");
        return -1;
    }
    
    if ((argsptr->altflags & FUSE_MOPT_AUTO_XATTR) &&
        (argsptr->altflags & FUSE_MOPT_NATIVE_XATTR)) {
        fprintf(stderr,
             "'auto_xattr' can't be used with 'native_xattr'\n");
        return -1;
    }
    
    if (daemon_timeout < FUSE_MIN_DAEMON_TIMEOUT) {
        daemon_timeout = FUSE_MIN_DAEMON_TIMEOUT;
    }
    
    if (daemon_timeout > FUSE_MAX_DAEMON_TIMEOUT) {
        daemon_timeout = FUSE_MAX_DAEMON_TIMEOUT;
    }
    
    result = ioctl(fd, FUSEDEVIOCGETRANDOM, &drandom);
    if (result) {
        fprintf(stderr, "failed to negotiate with /dev/"
                OSXFUSE_DEVICE_BASENAME "%d\n", dindex);
        return -1;
    }
    
    argsptr->blocksize      = (uint32_t)blocksize;
    argsptr->daemon_timeout = (uint32_t)daemon_timeout;
    argsptr->fsid           = (uint32_t)fsid;
    argsptr->fssubtype      = (uint32_t)fssubtype;
    argsptr->iosize         = (uint32_t)iosize;
    argsptr->random         = drandom;
    
    char *daemon_name = NULL;
    char *daemon_path = getenv("_FUSE_DAEMON_PATH");
    if (daemon_path) {
        daemon_name = basename(daemon_path);
    }
    
    if (!fsname) {
        if (daemon_name) {
            snprintf(argsptr->fsname, MAXPATHLEN, "%s@" OSXFUSE_DEVICE_BASENAME
                     "%d", daemon_name, dindex);
        } else {
            snprintf(argsptr->fsname, MAXPATHLEN, "instance@"
                     OSXFUSE_DEVICE_BASENAME "%d", dindex);
        }
    } else {
        snprintf(argsptr->fsname, MAXPATHLEN, "%s", fsname);
    }
    
    if (fstypename) {
        if (strlen(fstypename) > FUSE_TYPE_NAME_MAXLEN) {
            fprintf(stderr, "fstypename can be at most %lu characters\n",
                 (long unsigned int) FUSE_TYPE_NAME_MAXLEN);
            return -1;
        } else {
            snprintf(argsptr->fstypename, MFSTYPENAMELEN, "%s", fstypename);
        }
    }
    
    if (!volname) {
        #pragma clang diagnostic push
        #pragma clang diagnostic ignored "-Wformat-extra-args"
        
        if (daemon_name) {
            snprintf(argsptr->volname, MAXPATHLEN, OSXFUSE_VOLNAME_DAEMON_FORMAT,
                     dindex, daemon_name);
        } else {
            snprintf(argsptr->volname, MAXPATHLEN, OSXFUSE_VOLNAME_FORMAT, dindex);
        }

        #pragma clang diagnostic pop
    } else {
        snprintf(argsptr->volname, MAXPATHLEN, "%s", volname);
    }
    
    /* Finally! */
    result = mount(OSXFUSE_NAME, argsptr->mntpath, mntflags, (void *)argsptr);
    
    if (result < 0) {
        result = errno;
        fprintf(stderr, "failed to mount %s@/dev/" OSXFUSE_DEVICE_BASENAME "%d\n",
                argsptr->mntpath, dindex);
    } else {
        const char *dict[][2] = { { kFUSEMountPathKey, argsptr->mntpath } };
        post_notification(NOTIFICATION_MOUNT, dict, 1);
    }
    
    return result;
}

// We will be called as follows by the FUSE library:
//
//   mount_macfuse -o OPTIONS... -q <fdnam> <mountpoint>

/* NX Changes */
int
mounter_main(int argc, char **argv)
/* NX */
{
    int retval = 0;
    
    int mntflags = 0;
    char *mntpath;
    int cfd;
    int cvers;

    int fd;
    int dindex;
    int rdev;
    fuse_mount_args args;
    
    memset((void *)&args, 0, sizeof(args));
    
    retval = ParseArguments(argc, argv, &mntflags, &args.altflags, &mntpath, &cfd, &cvers);
    if (retval) {
        return -1;
    }

    if (!getenv("_FUSE_CALL_BY_LIB") && !install_only && !load_only) {
        showhelp();
    }

    /* NX Changes */
    /*
     * Create the AuthorizationRef that we'll use through this application. We
     * ignore any error from this. A failure from AuthorizationCreate is very
     * unusual, and if it happens there's no way to recover; Authorization
     * Services just won't work.
     */
    /*
    retval = AuthorizationCreate(NULL, kAuthorizationEmptyEnvironment, kAuthorizationFlagDefaults, &gAuth);
    assert(retval == noErr && gAuth != NULL);
    */
    
    /*
     * For each of our commands, check to see if a right specification exists
     * and, if not, create it.
     *
     * The last parameter is the name of a ".strings" file that contains the
     * localised prompts for any custom rights that we use.
     */
    /*
    BASSetDefaultRules(gAuth,
                       kMounterCommandSet,
                       CFBundleGetIdentifier(CFBundleGetMainBundle()),
                       CFSTR("AuthorizationPrompts"));
    */

    // Install or update helper tool if necessary
    //retval = InstallHelperTool();
    /* NX */
    if (retval) {
        return retval;
    }

    if (install_only) {
        return 0;
    }
    
    /* NX Changes */
    // We load the kext by running nxfs.sh script.
    /*
    // Load file system kernel extension
    retval = LoadFileSystemKext();
    if (retval) {
        return retval;
    }
    */

    retval = check_kext_status();
    switch (retval) {
        case 0:
            break;
            
        case ESRCH:
            fprintf(stderr, "the " OSXFUSE_DISPLAY_NAME
                 " kernel extension is not loaded\n");
            return -1;
            
        case EINVAL:
            fprintf(stderr, "the loaded " OSXFUSE_DISPLAY_NAME
                 " kernel extension has a mismatched version\n");
            return -1;
            
        default:
            fprintf(stderr, "failed to query the loaded " OSXFUSE_DISPLAY_NAME
                 " kernel extension (%d)\n", retval);
            return -1;
    }

    if (load_only) {
        return 0;
    }
    
    // Open FUSE device
    retval = OpenDevice(&fd, &dindex, &rdev);
    if (retval) {
        return retval;
    }
    
    signal_fd = fd;
    atexit(&signal_idx_atexit_handler);

    retval = CreateMountPath(mntpath);
    if (retval) {
        return retval;
    }

    realpath(mntpath, args.mntpath);
    
    retval = send_fd(cfd, fd);
    if (retval == -1) {
        fprintf(stderr, "failed to send file descriptor");
        return -1;
    }
    
    args.rdev = rdev;
    retval = MountFileSystem(fd, dindex, &args, mntflags);
    
    if (cvers == 2) {
        send_mount_status(cfd, retval);
    }
    
    if (retval) {
        return retval;
    }
    
    signal_fd = -1;
    return retval;
}
